/*-------------------------------------------------------------------*/
/* CBPNN.C           Pasi Frnti & Olli Virmajoki                    */
/*                                                                   */
/* Fast exact Pairwise Nearest Neighbour method 					      */
/*                                                                   */
/*-------------------------------------------------------------------*/

#define ProgName       "CBPNN"
#define VersionNumber  "Version 0.19"
#define LastUpdated    "19.09.04"  /*OV*/

/* ----------------------------------------------------------------- */

#include <assert.h>
#include <float.h>
#include <time.h>
#include <values.h>

#define  FACTFILE  "cbpnn.fac" 

#include "pnnfast.h"

#include "parametr.c"
#include "cb.h"
#include "file.h"
#include "memctrl.h"
#include "random.h"
#include "solution.h"
#include "sortcb.h"

/*--------------------------- Constants -----------------------------*/

#define  MAXFILENAME    64
#define  InitialSolutionGiven   (InitCBName[0] != 0x00)


/* =================  T I M E   F U N C T I O N S  ================= */


static long SetWatch(long* start)
{
  time_t tmp;
  return( *start = (long)time(&tmp) );
}


/* ----------------------------------------------------------------- */


static void PrintTime(long start)
{
  time_t tmp;
  long t = (long)time(&tmp) - start;
  long tout,h,m,s;

  tout = t;
  s = t % 60;
  t = t / 60;
  m = t % 60;
  h = t / 60;

  switch( Value(QuietLevel) )
    {
    case 0:  ;
    case 1:  printf(" %li\n",tout);
             break;
    case 2:
    case 3:
    case 4:  printf("  %li:%02li:%02li\n", h, m, s);
             break;
    default: printf("Time: %li:%02li:%02li\n", h, m, s);
             break;
    }
}


/*============================  P R I N T  ===============================*/


void PrintInfo(void)
{
  printf("\n%s %s %s\n", ProgName, VersionNumber, LastUpdated);
  printf("CBCROSS - crosses two input solutions using Genetic Algorithm.\n");
  printf("Usage: %s [%coption] <training set> <solution 1> <solution 2>\n", ProgName, OPTION_SYMBOL);
  printf("For example: %s bridge solution1 solution2 newsolution \n", ProgName);
  printf("\n  Options:\n");
  PrintOptions();
  printf("\n%s %s %s\n", ProgName, VersionNumber, LastUpdated);
}


/*-------------------------------------------------------------------*/


static void PrintOperatingInfo(char* TSName,
                               char* InitCBName,
                               char* CBName,
                               char* PAName)
{
  if( Value(QuietLevel) >= 2 )
    {
    printf("\n%s %s %s\n", ProgName, VersionNumber, LastUpdated);
	printf("Training Set:        %s\n", TSName);
    if(InitialSolutionGiven)
    printf("Initial codebook:    %s\n", InitCBName);
    printf("Output codebook:     %s\n", CBName);
    if(Value(SavePartition))
    printf("Output partition:    %s\n", PAName);
    printf("\n");
    PrintSelectedOptions();
    printf("\n");
    }
}


/*-------------------------------------------------------------------*/


static void PrintResult(long watch, double error)
{
  switch( Value(QuietLevel) )
	 {
     case 0:  break;
     case 1:
     case 2:  printf("%9.4f  ", error); break;
     default: printf("Distortion: %-9.4f\n", error); break;
	 }
  PrintTime(watch);
}


/*======================  SOLUTION HANDLING  ========================*/


static void GenerateInitialSolution(char*		  InitCBName,
                                    TRAININGSET*  TS,
                                    CODEBOOK*     CB,
                                    PARTITIONING* P)
/* Currently initial solution can only be CODEBOOK. It should be
   possible to input also partition as an input solution. */
{
  if( InitialSolutionGiven )
	 {
	 ReadCodebook(InitCBName, CB);
	 CreateNewPartitioning(P, TS, BookSize(CB));
	 GenerateOptimalPartitioning(TS, CB, P);
	 AddGenerationMethod(CB, "PNN from init");
	 }
  else	/* Generate initial codebook from TS */
    {
	CreateNewCodebook(CB, BookSize(TS), TS);
    CopyCodebook(TS, CB);
    SetVersionString(CB, CBFILE);
	CreateNewPartitioning(P, TS, BookSize(CB));
    PutAllInOwnPartition(TS, P);
    AddGenerationMethod(CB, "PNN");
  }
}


/*-------------------------------------------------------------------*/


static void SaveSolution(char*         CBName,
                         char*         PAName,
                         TRAININGSET*  TS,
                         CODEBOOK*     CB,
                         PARTITIONING* P)
{
  int ndups;

  ndups = DuplicatesInCodebook(CB);
  if( ndups>0 && Value(QuietLevel)>=2 )
    {
    printf("WARNING: Final CB contains %i duplicates.\n", ndups);
    }
  SortCodebook(CB, DATA_DESCENDING);
  WriteCodebook(CBName, CB, Value(OverWrite));
  if(Value(SavePartition))
     {
     WritePartitioning(PAName, P, TS, Value(OverWrite));
     }
}


/*-------------------------------------------------------------------*/


/*===========================  M A I N  ==============================*/


int main(int argc, char* argv[])
{
  TRAININGSET   TS;
  CODEBOOK      CB;
  PARTITIONING  P;
  char          TSName[MAXFILENAME]      = { '\0' };
  char          InitCBName[MAXFILENAME]  = { '\0' };
  char          CBName[MAXFILENAME]      = { '\0' };
  char          PAName[MAXFILENAME]      = { '\0' };
  double        Error=0.0;
  long          watch;
  
  int i;

  ParameterInfo paraminfo[3] = { { TSName,     FormatNameTS, 0, INFILE  },
                                 { InitCBName, FormatNameCB, 1, INFILE  },
								 { CBName,	   FormatNameCB, 0, OUTFILE } };

  ParseParameters(argc, argv, 3, paraminfo);
  PickFileName(CBName, PAName);
  CheckFileName(PAName, FormatNamePA);
  PrintOperatingInfo(TSName, InitCBName, CBName, PAName);
  ReadTrainingSet(TSName, &TS);

  SetWatch(&watch);

  /*for(i=0; i<100; i++){*/
  GenerateInitialSolution(InitCBName, &TS, &CB, &P);

  switch( Value(PNNMethod) )
  {
  case PNN:
      PNNFast(&TS, &CB, &P, Value(CodebookSize), Value(QuietLevel), watch,
	  Value(Pds), Value(Mps), Value(Lazy));
      break;
  case OPT1:
      printf("\nBrute force \n\n");
      Error = PNNGOp1(&TS, &CB, &P, Value(CodebookSize), Value(QuietLevel), watch,
      Value(Pds), Value(Mps), Value(Lazy)); 
      break;
  case OPT2:
      //printf("\nBranch and bound / Trivial termination \n\n");
      Error = PNNGOp2(&TS, &CB, &P, Value(CodebookSize), Value(QuietLevel), watch,
      Value(Pds), Value(Mps), Value(Lazy));
      break;
  case OPT3:
      printf("\nBranch and bound / Better termination \n\n");
      Error = PNNGOp3(&TS, &CB, &P, Value(CodebookSize), Value(QuietLevel), watch,
      Value(Pds), Value(Mps), Value(Lazy)); 
      break;
  case PIWI:
      printf("\nBranch and bound / Piecewise optimization \n\n");
      Error = PNNGBb2(&TS, &CB, &P, Value(CodebookSize), Value(StepSize), Value(QuietLevel), watch,
      Value(Pds), Value(Mps), Value(Lazy)); 
      break;
  case LOAH:
      printf("\nBranch and bound / Look-ahead optimization \n\n");
      Error = PNNGBb3(&TS, &CB, &P, Value(CodebookSize), Value(StepSize), Value(QuietLevel), watch,
      Value(Pds), Value(Mps), Value(Lazy));
      break;
  case DUFRSITE:
      printf("\nBranch and bound / Duplicate free; simple termination \n\n");
      Error = PNNGBb4(&TS, &CB, &P, Value(CodebookSize), Value(QuietLevel), watch,
      Value(Pds), Value(Mps), Value(Lazy));
      break;
  case DUFRNOTE:
      printf("\nBranch and bound / Duplicate free; no termination \n\n");
      Error = PNNGBb5(&TS, &CB, &P, Value(CodebookSize), Value(QuietLevel), watch,
      Value(Pds), Value(Mps), Value(Lazy));         
      break;
  case DUFRSITEWOPA:
      printf("\nBranch and bound / Duplicate free; simple termination without partition \n\n");
      Error = PNNGBb4_2(&TS, &CB, &P, Value(CodebookSize), Value(QuietLevel), watch,
      Value(Pds), Value(Mps), Value(Lazy));
      break;
  case DUFRNOTEWOPA:
      printf("\nBranch and bound / Duplicate free; no termination without partition \n\n");
      Error = PNNGBb5_2(&TS, &CB, &P, Value(CodebookSize), Value(QuietLevel), watch,
      Value(Pds), Value(Mps), Value(Lazy));
      break;
  default:
      printf("\nYou may have a wrong PNNMethod option !\nAllowed options are 0 - 8!\n\n");
      break;
  }

  Error = AverageErrorCBFast(&TS, &CB, &P, MSE);
  
  PrintResult(watch, PrintableError(Error,&CB));
  SaveSolution(CBName, PAName, &TS, &CB, &P);
  FreeCodebook(&CB);
  FreePartitioning(&P);
  /*}*/
  FreeCodebook(&TS);
  checkmemory();
  return( 0 );
}

