/*-------------------------------------------------------------------*/
/* PNNGBB5.C          Pasi Frnti & Olli Virmajoki                   */
/*                    & Timo Kaukoranta                              */
/*                                                                   */
/* Fast PNN method / Global Optimization                             */
/*   Branch and Bound / Variant duplicate free ; no termination      */
/*                                                                   */
/*                                                                   */
/*-------------------------------------------------------------------*/

#define ProgName       "PNN-GBB5"
#define VersionNumber  "Version 0.5"
#define LastUpdated    "19.9.2004"

/*-------------------------------------------------------------------*/

#include <assert.h>
#include <float.h>
#include <stdio.h>
#include <time.h>
#include <values.h>

#include "cb.h"
#include "interfc.h"
#include "memctrl.h"
#include "pnnfast.h"


/*---------------------------  M i s c  -----------------------------*/


/*#define  STATISTICS 1*/
#ifdef STATISTICS
#define STATCODE(a)a
#else
#define STATCODE(a)
#endif

#define  round(a)       ((a)+0.5)
#define  roundpds(a)    ((a)+0.9999999)
#define  swap(a,b)      { int t = a; a = b; b = t; }

static llong cfsol=0LL;
static llong csol=0LL;


/*===================================================================*/


static void PrintTime(long start)
{
  time_t tmp;
  long t = (long)time(&tmp) - start;
  long h,m,s;

  s=t%60; t=t/60; m=t%60; h=t/60;
  printf("  %li:%02li:%02li\n", h, m, s);
  fflush(stdout);
}


/* ========================  OPT  ================================== */


static void CreateDuplicateState(TRAININGSET*   TS,
                                 CODEBOOK*      CB,
                                 PARTITIONING*  P,
                                 CODEBOOK*      CB2,
                                 PARTITIONING*  P2)
{
    CreateNewCodebook(CB2, BookSize(CB), TS);
    CreateNewPartitioning(P2, TS, BookSize(CB));
    CopyCodebook(CB, CB2);
    CopyPartitioning(P, P2);
}


static void FreeDuplicateState(CODEBOOK*     CB2,
                               PARTITIONING* P2)
{
    FreeCodebook(CB2);
    FreePartitioning(P2);
}


static void CreateDuplicateSolution(TRAININGSET*    TS,
                                    CODEBOOK*       CB,
                                    PARTITIONING*   P,
                                    CODEBOOK*       CB2,
                                    PARTITIONING*   P2)
{
    CreateNewCodebook(CB2, BookSize(CB), TS);
    CreateNewPartitioning(P2, TS, BookSize(CB));
    CopyCodebook(CB, CB2);
    CopyPartitioning(P, P2);
}


static void FreeDuplicateSolution(CODEBOOK*    CB2,
                                  PARTITIONING* P2)
{
    FreeCodebook(CB2);
    FreePartitioning(P2);
}


static void Depthsearch(TRAININGSET*  TS,
                        CODEBOOK*     CB,
                        PARTITIONING* P,
                        int           FinalSize,
                        CODEBOOK*     CBOpt,
                        PARTITIONING* POpt,
                        double*       Error,
                        double*       ErrorOpt,
                        int           Quiet,
                        int           a,
                        int           b)
{
  int i, j;
  CODEBOOK     CB2;
  PARTITIONING P2;
  double Error0=0.;
  double* Error2=&Error0;

  /* If we have passed the last item, then move ahead.
     This index modification could be in recursive call instead.
     Or implemented by other way. */
  if( b == BookSize(CB) )
    {
    a++;
    b = a + 1;
    }
  for( i=a; i<FinalSize; i++ )
    {
    /* First loop has unique behaviour. */
    for( j = (i==a ? b : i+1); j<BookSize(CB); j++ )
      {
      CreateDuplicateState(TS, CB, P, &CB2, &P2);
      *Error2=*Error;

      *Error2 = *Error2 + (double)(VectorDist(Vector(&CB2,i), Vector(&CB2,j), 
          VectorSize(&CB2)))*
        ((double)VectorFreq(&CB2,i) * (double)VectorFreq(&CB2,j)) /
        ((double)VectorFreq(&CB2,i) + (double)VectorFreq(&CB2,j));

      /* Merge */
      JoinPartitions(TS, &P2, i, j);
      PartitionCentroid(&P2, i, &Node(&CB2, i));
      CopyNode(&Node(&CB2, BookSize(&CB2)-1), &Node(&CB2, j), VectorSize(&CB2));
      BookSize(&CB2)--;
      csol+=1;

      /* Codebook with right size reached. */
      if( BookSize(&CB2) == FinalSize )
        {
        //Error = AverageErrorCBFast(TS, &CB2, &P2, MSE);

        if( *Error2/(double)(TotalFreq(TS)*VectorSize(TS)) < *ErrorOpt )
          {
          *ErrorOpt = *Error2/(double)(TotalFreq(TS)*VectorSize(TS));
          CopyCodebook(&CB2, CBOpt);
          CopyPartitioning(&P2, POpt);
          }
        cfsol+=1;

        if(Quiet >= 2)
          {
          int zz;
          printf("P of %4.0f : ", (double)cfsol);
          for( zz=0; zz < BookSize(TS); zz++ )
            {
            printf("%3i ", Map(&P2, zz));
            }
          printf("\n");
          }
        }
      else
        {
        Depthsearch(TS, &CB2, &P2, FinalSize, CBOpt, POpt, Error2, ErrorOpt, Quiet, i, j);
        }

      FreeDuplicateState(&CB2, &P2);

      }
    }
}


/* ================================================================= */


double PNNGBb5(TRAININGSET*   TS,
               CODEBOOK*      CB,
               PARTITIONING*  P,
               int            FinalSize,
               int            Quiet,
               long           watch,
               int            Pds,
               int            Mps,
               int            Lazy)
{
  CODEBOOK      CBOpt;
  PARTITIONING  POpt;
  double        ErrorOpt;
  double Error0=.0;
  double *Error = &Error0;

  if( BookSize(CB) > FinalSize )
    {
    CreateDuplicateSolution(TS, CB, P, &CBOpt, &POpt);
    PNNFast(TS, &CBOpt, &POpt, FinalSize, Quiet, watch, Pds, Mps, Lazy);
    ErrorOpt = AverageErrorCBFast(TS, &CBOpt, &POpt, MSE);
    printf("\n%7.4f",ErrorOpt);
    PrintTime(watch);

    RemoveEmptyPartitions(TS, CB, P, FinalSize);

    Depthsearch(TS, CB, P, FinalSize, &CBOpt, &POpt, Error, &ErrorOpt, Quiet, 0, 1);

    CopyCodebook(&CBOpt, CB);
    CopyPartitioning(&POpt, P);

    FreeDuplicateSolution(&CBOpt, &POpt);
    }
  else
    {
    ErrorOpt = AverageErrorCBFast(TS, CB, P, MSE);
    }
  printf("Final solutions: %7.0f All solutions: %7.0f \n",(double)cfsol,(double)csol);
  return( ErrorOpt );
}


/* ================================================================= */


static void CentroidOfTwoVectors(CODEBOOK* CB, BOOKNODE* cent, int p1, int p2)
{
  int k;
  int freqsum;

  freqsum = VectorFreq(CB,p1) + VectorFreq(CB,p2);

  for( k=0; k<VectorSize(CB); k++)
    {
	if( freqsum > 0 )
	cent->vector[k] = round( (double)(VectorFreq(CB, p1)*VectorScalar(CB, p1, k) +
                                      VectorFreq(CB, p2)*VectorScalar(CB, p2, k))
                             / (double)freqsum );
	else cent->vector[k] = 0;
    }
  cent->freq = freqsum;
}


static void Depthsearch2(CODEBOOK*    CB,
                        int           FinalSize,
                        CODEBOOK*     CBOpt,
                        double*       Error,
                        double*       ErrorOpt,
                        double        Tfvs,
                        int           a,
                        int           b)
{
  int i, j;
  CODEBOOK     CB2;
  double Error0=0.;
  double* Error2=&Error0;

  /* If we have passed the last item, then move ahead.
     This index modification could be in recursive call instead.
     Or implemented by other way. */
  if( b == BookSize(CB) )
    {
    a++;
    b = a + 1;
    }
  for( i=a; i<FinalSize; i++ )
    {
    /* First loop has unique behaviour. */
    for( j = (i==a ? b : i+1); j<BookSize(CB); j++ )
      {
      CreateNewCodebook(&CB2, BookSize(CB), CB);
      CopyCodebook(CB, &CB2);
      *Error2=*Error;

      *Error2 = *Error2 + (double)(VectorDist(Vector(&CB2,i), Vector(&CB2,j), 
          VectorSize(&CB2)))*
        ((double)VectorFreq(&CB2,i) * (double)VectorFreq(&CB2,j)) /
        ((double)VectorFreq(&CB2,i) + (double)VectorFreq(&CB2,j));

      /* Merge */
      CentroidOfTwoVectors(&CB2, &Node(&CB2, i), i, j); 
      CopyNode(&Node(&CB2, BookSize(&CB2)-1), &Node(&CB2, j), VectorSize(&CB2));
      BookSize(&CB2)--;
      csol+=1;

      if( BookSize(&CB2) == FinalSize )
        {
        if( *Error2/Tfvs < *ErrorOpt )
          {
          *ErrorOpt = *Error2/Tfvs;
          CopyCodebook(&CB2, CBOpt);
          }
        cfsol+=1;
        }
      else
        {
        Depthsearch2(&CB2, FinalSize, CBOpt, Error2, ErrorOpt, Tfvs, i, j);
        }
      FreeCodebook(&CB2);
      }
    }
}


double PNNGBb5_2(TRAININGSET* TS,
               CODEBOOK*      CB,
               PARTITIONING*  P,
               int            FinalSize,
               int            Quiet,
               long           watch,
               int            Pds,
               int            Mps,
               int            Lazy)
{
  CODEBOOK      CBOpt;
  PARTITIONING  POpt;
  double        ErrorOpt;
  double Error0=.0;
  double *Error = &Error0;
  double Tfvs;

  if( BookSize(CB) > FinalSize )
    {
    CreateDuplicateSolution(TS, CB, P, &CBOpt, &POpt);

    PNNFast(TS, &CBOpt, &POpt, FinalSize, Quiet, watch, Pds, Mps, Lazy);
    ErrorOpt = AverageErrorCBFast(TS, &CBOpt, &POpt, MSE);
    printf("\n%7.4f",ErrorOpt);
    PrintTime(watch);

    Tfvs = (double)(TotalFreq(TS)*VectorSize(TS));

    Depthsearch2(CB, FinalSize, &CBOpt, Error, &ErrorOpt, Tfvs, 0, 1);

    CopyCodebook(&CBOpt, CB);
    CopyPartitioning(&POpt, P);

    FreeDuplicateSolution(&CBOpt, &POpt);
    }
  else
    {
    ErrorOpt = AverageErrorCBFast(TS, CB, P, MSE);
    }
  printf("Final solutions: %7.0f All solutions: %7.0f \n",(double)cfsol,(double)csol);
  return( ErrorOpt );
}


/* ================================================================= */