/*-------------------------------------------------------------------*/
/* PNNGOP2.C          Pasi Frnti & Olli Virmajoki                   */
/*                                                                   */
/* Fast PNN method / Global Optimization                             */
/*   Branch and bound / Backtrack programming                        */
/*	  Finds the Global Optimal Solution						         		*/
/*                                                                   */
/*-------------------------------------------------------------------*/

#define ProgName	   "PNN-GOPT"
#define VersionNumber  "Version 0.19"
#define LastUpdated    "7.10.2004"

/*-------------------------------------------------------------------*/

#include <assert.h>
#include <float.h>
#include <stdio.h>
#include <time.h>
#include <values.h>
#include "cb.h"
#include "interfc.h"
#include "memctrl.h"
#include "random.h"
#include "sort.h"
#include "solution.h"
#include "pnnfast.h"


/*---------------------------  M i s c	-----------------------------*/


#define  round(a)		((a)+0.5)
#define  roundpds(a)    ((a)+0.9999999)
#define  swap(a,b)      { int t = a; a = b; b = t; }


/*===================================================================*/


CODEBOOK CurrentCodebook;
int* BestPartitioning;
long** SumOfVectorsInPartition;
long* CountOfVectorsInPartition;


/*-------------------------------------------------------------------*/


static void PrintTime(long start)
{
  time_t tmp;
  long t = (long)time(&tmp) - start;
  long h,m,s;

  s=t%60; t=t/60; m=t%60; h=t/60;
  printf("  %li:%02li:%02li\n", h, m, s);
  fflush(stdout);
}


/*-------------------------------------------------------------------*/


static void VectorZero(int* vector, int dim)
{
  int k;
  for( k = 0; k < dim; k++ )
    {
    vector[k]=0;
    }
}


/*-------------------------------------------------------------------*/


void CalcVector(int* v, int* s1, int* s2, int N, int Vsize)
{
  int i;

  if( s1 == NULL || s2 == NULL )
    {
    ErrorMessage("ERROR: CalcVector source=%p source=%p\n", s1, s2);
    ExitProcessing( -1 );
    }
  for( i=0; i<Vsize; i++ )
    {
      if(N==0)
      {
          v[i]=s2[i];
      }
      else
      v[i] = round(  (double)( N*s1[i] + s2[i] ) / (double)(N+1) );
    }
}


/*-------------------------------------------------------------------*/


void CopyPartitioningOwn(int* dest, int source[], int BookSize)
{
    int i;
    for(i=0; i<BookSize; i++)
        dest[i]=source[i+1]-1;
}


/*-------------------------------------------------------------------*/


void PrintBestPartitioning(int* dest, int BookSize)
{
    int i;
    for(i=0; i<BookSize; i++)
        printf("\n %d -> %d",i, dest[i]);
}


/*--------------------------------------------------------------------*/


void PartitioningToCodeVectors(TRAININGSET* TS, int* p, CODEBOOK *cb,
										int FinalSize)
/* calculates codevectors from partitioning p and saves them to cb */

{
int i,j;
for (i=0; i<FinalSize; i++)
    {
    CountOfVectorsInPartition[i]=0;
    for (j=0; j<VectorSize(TS); j++)
        SumOfVectorsInPartition[i][j]=0L;
    }
for (i=0; i<BookSize(TS); i++)
    {
    (CountOfVectorsInPartition[p[i]])+=VectorFreq(TS,i);//WW
    //printf("\ni= %d Vf= %d p= %d", i ,VectorFreq(TS,i), p[i]);
    for (j=0; j<VectorSize(TS); j++)
        SumOfVectorsInPartition[p[i]][j] +=
            VectorScalar(TS,i,j)*VectorFreq(TS,i);
                    /* add vectors to corresponding partition sum */
    }
for (i=0; i< FinalSize; i++)
    {
    if (CountOfVectorsInPartition[i]>0)
        for (j=0; j<VectorSize(TS); j++)
            VectorScalar(cb,i,j)=(BYTE) ((double)SumOfVectorsInPartition[i][j] /
                                 (double)CountOfVectorsInPartition[i]+0.5);
    else
    {
        printf("\nProblems");
        CopyVector(Vector(cb,i),Vector(TS,0), VectorSize(TS));
    }
    }
}


/*-------------------------------------------------------------------*/


double PartitioningError(int* p, TRAININGSET* TS, int FinalSize)
/* Calculates the mean error of partitioning p.
   Destroys the contents of CurrentCodebook. */
{
int i;
long total=0;
PartitioningToCodeVectors(TS, p, &CurrentCodebook, FinalSize);
for (i=0; i<BookSize(TS); i++)
    total+=VectorDist(Vector(TS,i),Vector(&CurrentCodebook,p[i]),
                      VectorSize(TS)) * VectorFreq(TS,i);
return ( (double)total / (double)((double)(TS->TotalFreq)*VectorSize(TS)));
}


/* ================================================================= */


void Initialize(TRAININGSET* TS, VECTORTYPE** Ct, int FinalSize)
{
	int i;
   int VectorSize=VectorSize(TS);
   int BookSize = BookSize(TS);
	*Ct = (VECTORTYPE**)allocate(VectorSize*sizeof(VECTORTYPE*));
   for(i=0; i<BookSize+1; i++)
   	*Ct[i] = (VECTORTYPE*)allocate(VectorSize*sizeof(VECTORELEMENT*));
   BestPartitioning = (int*)allocate(sizeof(int)*BookSize);
  	CreateNewCodebook(&CurrentCodebook, FinalSize, TS);
  	CountOfVectorsInPartition =(long*)allocate(sizeof(long)*FinalSize);
   SumOfVectorsInPartition = (long**)allocate(sizeof(long*)*FinalSize);
  	for(i=0; i<FinalSize; i++)
      SumOfVectorsInPartition[i]=(long*)allocate(sizeof(long)*VectorSize);
}


void ShutDown(VECTORTYPE* Ct, int BookSize, int FinalSize)
{
	int i;
	for(i=0; i<BookSize+1; i++)
		deallocate(&Ct[i]);
   deallocate(Ct);
   deallocate (BestPartitioning);
   FreeCodebook(&CurrentCodebook);
   for(i=0;i<FinalSize;i++)
      deallocate(SumOfVectorsInPartition[i]);
   deallocate(SumOfVectorsInPartition);
   deallocate(CountOfVectorsInPartition);
}


/* ================================================================= */


double PNNGOp2(TRAININGSET*	  TS,
			      CODEBOOK*		  CB,
			      PARTITIONING*    P,
               int              FinalSize,
               int              Quiet,
               long             watch,
               int              Pds,
               int              Mps,
               int              Lazy)
{
  double ErrorIni = MAXLLONG, BestError=0.;
  double *ErrorOpt = &ErrorIni;
  double factor=0.,Tfvs=0.;
  llong min, mine, b[100], B;
  int w[100], p[100];
  int i,j=0,j2,k,k1,l=0,var=0,M=0, BS=0;
  int fii[100][50];
  int N[100][50];
  VECTORTYPE* Ct;

  printf("\n\nBranch and bound : BackTrack Programming \n");
  Initialize(TS, &Ct, FinalSize);

  if(BookSize(CB) > FinalSize)
  {
  printf("\nPNN solution:");
  PNNFast(TS, CB, P, FinalSize, 0, watch, Pds, Mps, Lazy);
  (*ErrorOpt) = AverageErrorCBFast(TS, CB, P, MSE);
  printf("\n%7.4f",*ErrorOpt);
  }
  PrintTime(watch);


  printf("\n\nBegin of BackTrack \n");


  //Step1: /*Initialization*/
  /*************************/

	if(Quiet>1)
  		printf("\nStep1: Initialization \n");

  B = MAXLLONG;
  M = FinalSize;
  BS = BookSize(TS);
  Tfvs=(double)TotalFreq(TS)*(double)VectorSize(TS);
  for(i=0;i<BS+1;i++)
  {
  		w[i]=0;
      p[i]=0;
      b[i]=0LL;
      for(j=0;j<M+1;j++)
      {
      	fii[i][j]=0;
      	N[i][j]=0;
         VectorZero(&Ct[i][j], VectorSize(TS));
      }
  }
  k=1;

/*  mine = VectorDist(Vector(TS,0), Vector(TS,1), VectorSize(TS));
  printf("\n test01= %7.4f",(double)mine/2.);
  mine = VectorDist(Vector(TS,0), Vector(TS,2), VectorSize(TS));
  printf("\n test02= %7.4f",(double)mine/2.);
  mine = VectorDist(Vector(TS,1), Vector(TS,2), VectorSize(TS));
  printf("\n test12= %7.4f",(double)mine/2.);
*/


  Step2: /*Partial Classification Generation*/
  /******************************************/

  if(Quiet>1)
  		printf("\nStep2 k = %d ",k);

  min=MAXLLONG; mine=MAXLLONG;
  for(j2=1; j2<M+1; j2++)
  {
  		if(fii[k][j2]==0)
    	{
      	//printf("\nN= %d",N[k-1][j2]);
        	factor = (double)(N[k-1][j2]) / (double)(N[k-1][j2]+1);
        	//printf("\nfactor= %f",factor);

        	mine = round(factor * (double)(VectorDist(Vector(TS,k-1), &Ct[k-1][j2],
            VectorSize(TS))) );
        	if(Quiet>1)
        		printf("\n mine= %7.4f",(double)mine);
        	if(mine<min)
        	{
        		min=mine;
        		j=j2;
        		w[k]=j2;
        	}
    	}
  }
  b[k]=b[k-1]+min;
  if(Quiet>1)
  		{
  		printf("\n j= %d b[%d]= %f  \n",j, k, (double)b[k]);
  		for(k1=1;k1<k+1;k1++)
  		printf(" w[%d]= %d",k1,w[k1]);
  		}


  //Step3: /*Evaluation*/
  /*********************/

  if(b[k]<B)
  		goto Step4;
  else
  		goto Step5;


  Step4: /*Forward Step*/
  /*********************/

  if(Quiet>1)
  		printf("\nStep4 k = %d ",k);

  if(k==BS)
  		goto Step7;
  else
  {
  		fii[k][w[k]]=1;
      for(l=1;l<M+1;l++)
  		{
      if(j!=l)
      {
        N[k][l]=N[k-1][l];
        CopyVector(&Ct[k-1][l], &Ct[k][l], VectorSize(TS));
      }
      else /* j==l */
      {
        N[k][l]=N[k-1][l]+1;
        if(N[k][l]==1)
        		CopyVector(Vector(TS,k-1), &Ct[k][l], VectorSize(TS));
        else
	    		CalcVector(&Ct[k][l], &Ct[k-1][l], Vector(TS,k-1), N[k-1][l],
            VectorSize(TS));
      }
  		}
  		if(Quiet>1)
  			for(l=1;l<M+1;l++)printf(" N[%d][%d]= %d",k,l,N[k][l]);
      k++;
  		goto Step2;
  }


  Step5: /*Backward Step*/
  /**********************/

  for(j=1;j<M+1;j++)
      fii[k][j]=0;
  k--;
  if(k==0)
  		goto Stepf;
  else
  		goto Step6;


  Step6: /*Seek New Classification at Level k*/
  /*******************************************/

  var=1;
  for(l=1;l<M+1;l++)
  		if(fii[k][l]==0)
      	var=0;
  if(var)
  		goto Step5;
  else
  		goto Step2;


  Step7: /*Store Classification and Cost*/
  /**************************************/

  B=b[k];
  for(i=1; i<BS+1; i++)
      p[i]=w[i];
  CopyPartitioningOwn(BestPartitioning, w, BS);
  if(Quiet>1)
  		printf("\nStep7 B= %f \n",(double)B);
  goto Step5;


  Stepf: /*Final step*/
  /*******************/

  printf("\n\nFinal step: B= %7.4f ",(double)B);
  printf("\nB/Tsvs= %7.4f ",(double)B/Tfvs);
  for(i=1; i<BS+1;i++)
      printf("\n k= %d -> %d", i, p[i]);
  PrintBestPartitioning(BestPartitioning, BS);
  BestError= PartitioningError(BestPartitioning, TS, FinalSize);
  printf("\nError %f\n",BestError);


  ShutDown(Ct, BookSize(TS), FinalSize);
  printf("\nEnd of BackTrack\n\n");
  return (*ErrorOpt);
}

