/*-------------------------------------------------------------------*/
/* PNNGOP3.C          Pasi Frnti & Olli Virmajoki                   */
/*                                                                   */
/* Fast PNN method / Global Optimization                             */
/*   Branch and bound / Better termination                           */
/*	 Finds the Global Optimal Solution								 */
/*                                                                   */
/*-------------------------------------------------------------------*/

#define ProgName	   "PNN-GOP3"
#define VersionNumber  "Version 0.8"
#define LastUpdated    "16.9.2004"

/*-------------------------------------------------------------------*/

#include <assert.h>
#include <float.h>
#include <stdio.h>
#include <time.h>
#include <values.h>

#include "cb.h"
#include "interfc.h"
#include "memctrl.h"
#include "random.h"
#include "sort.h"
#include "solution.h"
#include "pnnfast.h"


/*---------------------------  M i s c	-----------------------------*/


/*#define  STATISTICS 1*/
#ifdef STATISTICS
#define STATCODE(a)a
#else
#define STATCODE(a)
#endif

#define  round(a)		((a)+0.5)
#define  roundpds(a)    ((a)+0.9999999)
#define  swap(a,b)      { int t = a; a = b; b = t; }

static llong cfsol=0LL;
static llong csol=0LL;


/*===================================================================*/


static void PrintTime(long start)
{
  time_t tmp;
  long t = (long)time(&tmp) - start;
  long h,m,s;

  s=t%60; t=t/60; m=t%60; h=t/60;
  printf("  %li:%02li:%02li\n", h, m, s);
  fflush(stdout);
}


/* ========================  OPT  ================================== */


static llong MergeDistortionO(CODEBOOK* CB, int p1, int p2)
{
  /* Original "inline" */
  int i=0, Vsize;
  llong   dist=0LL, diff, dist_limit=MAXLLONG;
  double  factor;

  Vsize = VectorSize(CB);
  factor = ((double)VectorFreq(CB,p1) * (double)VectorFreq(CB,p2)) /
           ((double)VectorFreq(CB,p1) + (double)VectorFreq(CB,p2));
	 /* Precondition: Vsize >= 1 */
     /* if dist >= dist_limit, returns at least factor * dist_limit */
  do
	{
	   diff = (llong)VectorScalar(CB,p1,i)-(llong)VectorScalar(CB,p2,i);
	   dist += diff * diff;
       STATCODE
           (
           ksum++; /* Statistical information */
           )
	} while(dist < dist_limit && ++i < Vsize);
  dist=round( factor * (double)dist);
  STATCODE
      (
      ndistcalcs++; /* Statistical information */
      )

  return( dist);
}


static void FindNearestClusterO(CODEBOOK* CB, int a, int* b,
								  llong* dist, int Pds)
{
  llong Dist;
  int i;

    (*dist)=MAXLLONG/10; (*b)=0;
    for( i=0; i<BookSize(CB); i++ )
    {
        if( a!=i)
	    {
		    Dist = MergeDistortionO(CB, a, i);
            if( Dist < (*dist) )
            {
                (*dist) = Dist;
                (*b)    = i;
            }
        }
    }
}


static void FindMergedPairO(CODEBOOK* CB, NEARESTTYPE NN[], int* a, int* b)
{
  llong MinDist=MAXLLONG;
  int i;

  for( i=0; i<BookSize(CB); i++ )
    {
       if( NN[i].VectorDist < MinDist )
	   {
          if(NN[i].Recalculate)
          {
              FindNearestClusterO(CB, i, &(NN[i].Vector), 
                &(NN[i].VectorDist), 0);
              NN[i].Recalculate = NO;
              if(NN[i].VectorDist < MinDist)
              {
                (*a) = i;
                (*b) = NN[i].Vector;
                MinDist = NN[i].VectorDist;
              }
          }
          else
          {
            (*a) = i;
            (*b) = NN[i].Vector;
            MinDist = NN[i].VectorDist;
          }
       }
    }
  if( (*a)>(*b) ) swap((*a),(*b));
}


static void CopyNN(NEARESTTYPE NNsource[], NEARESTTYPE NNdest[], int booksize)
{
    int i;
    for (i=0; i<booksize; i++)
    {
        NNdest[i].VectorDist = NNsource[i].VectorDist;
        NNdest[i].Vector = NNsource[i].Vector;
        NNdest[i].Recalculate = NNsource[i].Recalculate;
    }
}


static void CreateDuplicateState(TRAININGSET*   TS,
                                 CODEBOOK*      CB,
                                 PARTITIONING*  P,
                                 NEARESTTYPE    NN[],
                                 CODEBOOK*      CB2,
                                 PARTITIONING*  P2,
                                 NEARESTTYPE**  NN2)
{
    CreateNewCodebook(CB2, BookSize(CB), TS);
    CreateNewPartitioning(P2, TS, BookSize(CB));
    *NN2 = (NEARESTTYPE*) allocate(BookSize(CB) * sizeof(NEARESTTYPE));
    CopyCodebook(CB, CB2);
    CopyPartitioning(P, P2);
    CopyNN(NN, *NN2, BookSize(CB));
}


static void FreeDuplicateState(CODEBOOK*     CB2,
                               PARTITIONING* P2,
                               NEARESTTYPE   NN2[])
{
    FreeCodebook(CB2);
    FreePartitioning(P2);
    deallocate(NN2);
}


static void CreateDuplicateSolution(TRAININGSET*    TS,
                                    CODEBOOK*       CB,
                                    PARTITIONING*   P,
                                    CODEBOOK*       CB2,
                                    PARTITIONING*   P2)
{
    CreateNewCodebook(CB2, BookSize(CB), TS);
    CreateNewPartitioning(P2, TS, BookSize(CB));
    CopyCodebook(CB, CB2);
    CopyPartitioning(P, P2);
}


static void FreeDuplicateSolution(CODEBOOK*    CB2,
                                  PARTITIONING* P2)
{
    FreeCodebook(CB2);
    FreePartitioning(P2);
}


static void Depthsearch(TRAININGSET*  TS,
                        CODEBOOK*     CB,
                        PARTITIONING* P,
                        NEARESTTYPE   NN[],
                        int           FinalSize,
                        CODEBOOK*     CBOpt,
                        PARTITIONING* POpt,
                        double*		  Error,
                        double*       ErrorOpt,
                        int           Quiet)
{
    CODEBOOK CB2;
    PARTITIONING P2;
    NEARESTTYPE* NN2;
    int i, j, S;
    int a, b;
    double Error0=0.;
    double* Error2=&Error0;
    double ErrorL;

      S = BookSize(CB)-1;
      for( i=0; i < S; i++)
        for( j = i+1; j <= S; j++)
        {
          if(NN[i].Vector==j /*|| NN[j].Vector==i*/)
          {
            CreateDuplicateState(TS, CB, P, NN, &CB2, &P2, &NN2);
            *Error2=*Error;

            *Error2 = *Error2 + (double)(VectorDist(Vector(&CB2,i), Vector(&CB2,j), VectorSize(&CB2)))*
            ((double)VectorFreq(&CB2,i) * (double)VectorFreq(&CB2,j)) /
            ((double)VectorFreq(&CB2,i) + (double)VectorFreq(&CB2,j));

            MergeVectors(TS, &CB2, &P2, NN2, i, j);
            csol+=1;

            if(BookSize(&CB2) > FinalSize)
            {
                RecalculateDistances(&CB2, NN2, i, Quiet, 0, 0);
                FindMergedPairO(&CB2, NN2, &a, &b);

                ErrorL = *Error2/((double)TotalFreq(TS)*VectorSize(TS));
                ErrorL += (BookSize(&CB2)-FinalSize)*NN2[a].VectorDist/((double)TotalFreq(TS)*VectorSize(TS));

                if(ErrorL < *ErrorOpt)
                Depthsearch(TS, &CB2, &P2, NN2, FinalSize, CBOpt, POpt, Error2, ErrorOpt, Quiet);
            }
            else
            {
                cfsol+=1;
                if(*Error2/(double)(TotalFreq(TS)*VectorSize(TS)) < *ErrorOpt)
                {
                    *ErrorOpt = *Error2/(double)(TotalFreq(TS)*VectorSize(TS));
                    CopyCodebook(&CB2, CBOpt);
                    CopyPartitioning(&P2, POpt);
                }
                if(Quiet >= 2)
                printf("\nFinalSize= %d ",BookSize(&CB2));
            }
            FreeDuplicateState(&CB2, &P2, NN2);
            *Error2 = *Error;
          }
        }
}


/* ================================================================= */


double PNNGOp3(TRAININGSET*	  TS,
			   	CODEBOOK*		  CB,
			   	PARTITIONING*	  P,
               int              FinalSize,
               int              Quiet,
               long             watch,
               int              Pds,
               int              Mps,
               int              Lazy)
{
  CODEBOOK CBOpt, CB2;
  PARTITIONING POpt, P2;
  NEARESTTYPE*  NN;
  DATA D;
  double ErrorIni = MAXLLONG;
  double Error0=0.;
  double *Error=&Error0;
  double *ErrorOpt = &ErrorIni;

  if(BookSize(CB) > FinalSize)
  {
  CreateNewCodebook(&CBOpt, FinalSize, TS);
  CreateNewPartitioning(&POpt, TS, FinalSize);

  CreateDuplicateSolution(TS, CB, P, &CB2, &P2);
  PNNFast(TS, &CB2, &P2, FinalSize, Quiet, watch, Pds, Mps, Lazy);
  (*ErrorOpt) = AverageErrorCBFast(TS, &CB2, &P2, MSE);
  printf("\n%7.4f",*ErrorOpt);
  CopyCodebook(&CB2, &CBOpt);
  CopyPartitioning(&P2, &POpt);
  FreeDuplicateSolution(&CB2, &P2);
  PrintTime(watch);

  RemoveEmptyPartitions(TS, CB, P, FinalSize);

  CreateDataStructures(CB, &D, &NN, Mps);
  InitializeData(CB, &D, NN, Mps);
  CalculateAllDistances(CB, &D, NN, Quiet, watch, Pds, Mps);

  Depthsearch(TS, CB, P, NN, FinalSize, &CBOpt, &POpt, Error, ErrorOpt, Quiet);
  CopyCodebook(&CBOpt, CB);
  CopyPartitioning(&POpt, P);
  FreeCodebook(&CBOpt);
  FreePartitioning(&POpt);
  FreeDataStructures(&D, NN, Mps);
  }
  else
  {
    *ErrorOpt = AverageErrorCBFast(TS, CB, P, MSE);
  }
  printf("Final solutions: %7.0f All solutions: %7.0f \n",(double)cfsol,(double)csol);
  return (*ErrorOpt);
}

